(*******************************************************************)
(*  This is part of SymbolicKAT, it is distributed under the       *)
(*  terms of the GNU Lesser General Public License version 3       *)
(*           (see file LICENSE for more details)                   *)
(*                                                                 *)
(*  Copyright 2014: Damien Pous. (CNRS, LIP - ENS Lyon, UMR 5668)  *)
(*******************************************************************)

(** Binary Decision Diagrams (BDDs) *)

open Common

(** Multi-Terminal BDD nodes with leaves in ['a] *)
type 'a node

(** memories for BDDs nodes with leaves in ['a] (required to ensure unicity) *)
type 'a mem

(** empty memory, with given hash and equality functions for leaves *)
val init: ('a -> int) -> ('a -> 'a -> bool) -> 'a mem

(** constant BDD (leaf) *)
val constant: 'a mem -> 'a -> 'a node

(** applying a unary function to a BDD *)
val unary: 'b mem -> ('a -> 'b) -> 'a node -> 'b node

(** applying a binary function to two BDDs (the standard "apply" function) *)
val binary: 'c mem -> ('a -> 'b -> 'c) -> 'a node -> 'b node -> 'c node

(** type of BDD unifying functions *)
type 'a unifier = ((key * bool) list -> 'a -> 'a -> unit) -> 'a node -> 'a node -> unit

(** naive unifier where a set of pairs of nodes is stored. 
    The Boolean argument indicates whether to trace operations or not *)
val unify_naive: bool -> 'a unifier

(** union-find based unifier where a forest of pointers is stored.
    The Boolean argument indicates whether to trace operations or not *)
val unify_dsf: bool -> 'a unifier

(** identifier of a BDD node (unique w.r.t. a given memory) *)
val tag: 'a node -> int

(** hash of a BDD node *)
val hash: 'a node -> int

(** symbolic Boolean formula *)
type formula = bool node

(** the various connectives  *)
val neg: formula -> formula
val dsj: formula -> formula -> formula
val cnj: formula -> formula -> formula
val xor: formula -> formula -> formula
val iff: formula -> formula -> formula

(** the two constants *)
val bot: formula
val top: formula

(** literals: a single variable, or the negation of a variable *)
val var: key -> formula
val rav: key -> formula

(** witness assignation for a satisfiable formula *)
val witness: formula -> (key*bool) list

(** [times m z f n] multiplies the BDD [n] by the formula [f], using [z] where [f] is false *)
val times: 'a mem -> 'a node -> formula -> 'a node -> 'a node

(** pretty-printing a symbolic formula (the first argument specifies the current level, for parentheses) *)
val print_formula: int -> Format.formatter -> formula -> unit

(** tracing a BDD (for rendering) *)
val trace: ?exclude: 'a node -> (('a node -> unit) -> int -> 'a -> unit) -> 'a node -> unit


(** reset all memoisation caches *)
val reset_caches: unit -> unit

