(*******************************************************************)
(*  This is part of SymbolicKAT, it is distributed under the       *)
(*  terms of the GNU Lesser General Public License version 3       *)
(*           (see file LICENSE for more details)                   *)
(*                                                                 *)
(*  Copyright 2014: Damien Pous. (CNRS, LIP - ENS Lyon, UMR 5668)  *)
(*******************************************************************)

let get t x = try Some (Hashtbl.find t x) with Not_found -> None

module type QUEUE = sig
  type 'a t
  val empty: unit -> 'a t
  val singleton: 'a -> 'a t
  val push: 'a t -> 'a -> unit
  val pop: 'a t -> 'a option
  val fold : ('b -> 'a -> 'b) -> 'b -> 'a t -> 'b
end

module IntSet = Sets.NarithInlined
type int_set = IntSet.t

type 'a hval = 'a Hashcons.hash_consed
type 'a hset = 'a Hset.t
module Set = struct
  type 'a t = 'a list
  let empty = []
  let singleton x = [x]
  let union =
    let rec union h k = 
      match h,k with
	| [],l | l,[] -> l
	| x::h', y::k' -> match compare x y with
	    | 0 -> x::union h' k'
	    | 1 -> x::union h' k
	    | _ -> y::union h k'
    in union
  let rec mem x = function
    | [] -> false 
    | y::q -> match compare x y with
	| 1 -> false
	| 0 -> true
	| _ -> mem x q
  let add x = union [x]
  let fold = List.fold_right
  let iter = List.iter
  let map f = List.fold_left (fun x i -> union [f i] x) []
  let print g f x = 
    Format.pp_print_list ~pp_sep:(fun f () -> Format.pp_print_char f ',') g f x
end
type 'a set = 'a Set.t

type var = char
type key = char
type gstring_ = (var * (key*bool) list) list
type gstring = (key*bool) list * gstring_
let print_gstring f (alpha,l) =
  let rec atom = function
    | [] -> ()
    | (a,true)::q -> atom q; Format.fprintf f "%c" a
    | (a,false)::q -> atom q; Format.fprintf f "!%c" a
  in
  let rec list = function
    | [] -> ()
    | (p,alpha)::q -> list q; atom alpha; Format.fprintf f "%c" p
  in if l=[] && alpha=[] then Format.fprintf f "1" else (list l; atom alpha)

let paren h k x = if h>k then "("^^x^^")" else x

module Span = struct
  type 's t = (var*'s) list * 's
  let empty z = [],z
  let single a x z = [a,x],z
  let merge f =
    let rec merge h k = match h,k with
      | [],l | l,[] -> l
      | (a,x as ax)::h', (b,y as ay)::k' -> 
	match compare a b with
	  | 0 -> (a, f x y)::merge h' k'
	  | 1 -> ax::merge h' k
	  | _ -> ay::merge h k'
    in fun (h,z) (k,z') -> assert(z==z'); merge h k,z
  let map f (h,z) = List.map (fun (a,x) -> (a, f x)) h, f z
  let iter f (h,z) = List.iter (fun (a,x) -> f a x) h; f '_' z
  let iter2 f (h,z) (k,z') =
    assert(z==z');
    let rec iter2 h k = match h,k with
      | [],[] -> ()
      | [],(a,y)::k -> f a z y; iter2 [] k
      | (a,x)::h,[] -> f a x z; iter2 h []
      | (a,x)::h', (b,y)::k' -> 
	match compare a b with
	  | 0 -> f a x y; iter2 h' k'
	  | 1 -> f a x z; iter2 h' k
	  | _ -> f b z y; iter2 h  k'
    in iter2 h k
end
type 's span = 's Span.t
