(*******************************************************************)
(*  This is part of SymbolicKAT, it is distributed under the       *)
(*  terms of the GNU Lesser General Public License version 3       *)
(*           (see file LICENSE for more details)                   *)
(*                                                                 *)
(*  Copyright 2014: Damien Pous. (CNRS, LIP - ENS Lyon, UMR 5668)  *)
(*******************************************************************)

(** Common definitions and utilities *)

(** wrapper around [Hashtbl.find] *)
val get : ('a, 'b) Hashtbl.t -> 'a -> 'b option

(** signature for queues *)
module type QUEUE = sig
  type 'a t
  val empty : unit -> 'a t
  val singleton : 'a -> 'a t
  val push : 'a t -> 'a -> unit
  val pop : 'a t -> 'a option
  val fold : ('b -> 'a -> 'b) -> 'b -> 'a t -> 'b
end

(** efficient implementation of sets of integers *)
module IntSet: Sets.T
type int_set = IntSet.t

(** hash-consed values, ans sets of such values *)
type 'a hval = 'a Hashcons.hash_consed
type 'a hset = 'a Hset.t

(** (not so efficient) implementation of polymorphic sets *)
module Set: sig
  type 'a t
  val empty : 'a t
  val singleton : 'a -> 'a t
  val union : 'a t -> 'a t -> 'a t
  val mem : 'a -> 'a t -> bool
  val add : 'a -> 'a t -> 'a t
  val map : ('a -> 'b) -> 'a t -> 'b t
  val iter : ('a -> unit) -> 'a t -> unit
  val fold : ('a -> 'b -> 'b) -> 'a t -> 'b -> 'b
  val print: (Format.formatter -> 'a -> unit) -> Format.formatter -> 'a t -> unit
  end
type 'a set = 'a Set.t

(** the type of KAT atomic Kleene elements *)
type var = char

(** the type of KAT atomic tests *)
type key = char

(** unterminated guarded strings (for counter-examples) *)
type gstring_ = (var * (key * bool) list) list
(** guarded strings *)
type gstring = (key * bool) list * gstring_

(** pretty-printing for guarded strings *)
val print_gstring: Format.formatter -> gstring -> unit

(** utility for formatting parenthesised expressions *)
val paren: int -> int -> ('b, 'c, 'd, 'e, 'f, 'g) format6 -> ('b, 'c, 'd, 'e, 'f, 'g) format6

(** spans: sparse polymorphic maps indexed by [var] 
    these maps are used to represent explicit part of the
    transitions of guarded string automata *)
module Span: sig
  type 'a t
  (** [empty z] is the constant to [z] map *)
  val empty : 'a -> 'a t
  (** [single p x z] is map whose value is [x] on [p] and [z] everywhere else *)
  val single : var -> 'a -> 'a -> 'a t
  (** [merge f m n] applies [f] pointwise to [m] and [n], 
      assume that [m] and [z] share the same default value [z], 
      and that [f z z = z] *)
  val merge : ('a -> 'a -> 'a) -> 'a t -> 'a t -> 'a t
  (** pointwise application of a function  *)
  val map : ('a -> 'b) -> 'a t -> 'b t
  (** simple iterator *)
  val iter : (var -> 'a ->  unit) -> 'a t  -> unit
  (** [iter2 f m n] applies [f] to all non-trivial pairs obtained by zipping [m] and [n] *)
  val iter2 : (var -> 'a -> 'a -> unit) -> 'a t -> 'a t -> unit
end
type 's span = 's Span.t
