(*******************************************************************)
(*  This is part of SymbolicKAT, it is distributed under the       *)
(*  terms of the GNU Lesser General Public License version 3       *)
(*           (see file LICENSE for more details)                   *)
(*                                                                 *)
(*  Copyright 2014: Damien Pous. (CNRS, LIP - ENS Lyon, UMR 5668)  *)
(*******************************************************************)

open Common
open Automata

(* generic symbolic algorithm*)
module Symb = Equiv.Symbolic(Queues.BFS) 

(* quiet or not, stats or not *)
let quiet = ref !Sys.interactive
let stats = ref false

(* recording traces or not *)
let trace = ref false

(* do we use Hopcroft and Karp's version? *)
let hk = ref false

(* do we put expressions in strict star form first? *)
let ssf = ref false

(* which construction to use *)
let construction = ref `Antimirov
let set x () = construction := x

(* do we use up-to congruence *)
let congruence = ref false

(* hypotheses to exploit *)
let hypotheses = ref ""

let hexprs hyps x y = 
  let x,y,z = Hypotheses.eliminate hyps x y in
  if !ssf then Kat.ssf x, Kat.ssf y, Kat.ssf z else x,y,z

let hexpr x = let x,_,_ = hexprs (Parse.hyps !hypotheses) x (Kat.Tst Kat.Bot) in x

let equiv_ e a exclude x y = 
  let tracer = if !trace then (
    Trace.clear();
    let pp = SDFA.trace ~exclude a in
    pp y; pp x;
    Some (fun k x y ->
      let x = Bdd.tag (Bdd.constant a.SDFA.m x) in
      let y = Bdd.tag (Bdd.constant a.SDFA.m y) in
      match k with
	| `CE -> Trace.ce x y
	| `OK -> Trace.ok x y
	| `Skip -> Trace.skip x y
    ) 
  ) else None 
  in
  match 
    if !hk then e ?tracer (Bdd.unify_dsf !trace) a x y 
    else e ?tracer (Bdd.unify_naive !trace) a x y 
  with
    | None -> None
    | Some (x,y,w) -> 
      let b,c = a.SDFA.o x, a.SDFA.o y in
      Some (List.rev (Bdd.witness (Bdd.xor b c)), w)

(* Antimirov' construction *)
let antimirov e x y =
  let a,f = SNFA.reindex (Antimirov.nfa()) in
  equiv_ e (Determinisation.optimised a) IntSet.empty
    (f (Antimirov.split x)) (f (Antimirov.split y))

(* Brzozowski's construction *)
let brzozowski e x y = 
  equiv_ e (Brzozowski.dfa()) Kat.zer x y

(* Ilie and Yu's construction *)
let ilie_yu e x y =
  let a,x,y = IlieYu.enfa x y in
  let a = Determinisation.optimised (Epsilon.remove a) in
  equiv_ e a IntSet.empty (IntSet.singleton x) (IntSet.singleton y)

(* equivance check *)
let equiv x y = 
  match !construction with
    | `Antimirov -> antimirov (if !congruence then Symb.equiv_c else Symb.equiv) x y
    | `IlieYu -> ilie_yu (if !congruence then Symb.equiv_c else Symb.equiv) x y
    | `Brzozowski -> 
      if !congruence then failwith "up-to congruence cannot be used with Brzozowski's automaton"
      else brzozowski Symb.equiv x y

(* full comparison *)
let compare ?(hyps=[]) x y = 
  Stats.reset(); 
  let x',y',z' = hexprs hyps x y in
  (x',y'),
  match equiv x' y' with
    | None -> `E
    | Some w -> Stats.reset(); Trace.save(); match equiv z' y' with
	| None -> `L w
	| Some w1 -> Stats.reset(); match equiv z' x' with
	    | None -> `G w
	    | Some w2 -> Trace.restore(); `D (w1,w2)

(* exported equivance check *)
let equiv ?(hyps=[]) x y = 
  let x',y',_ = hexprs hyps x y in
  equiv x' y'

(* equivalence/inclusion/difference check *)
let check m ?(hyps="") x y =
  let x = Parse.expr x in
  let y = Parse.expr y in
  let hyps' = Parse.hyps hyps in
  match snd (compare ~hyps:hyps' x y),m with
    | `E,`E | `L _,`L | `G _,`G | `D _,`D -> ()
    | _ -> 
      if hyps="" then Format.eprintf "Invalid result on %a and %a\n%!" Kat.print_expr x Kat.print_expr y
      else Format.eprintf "Invalid result on %a and %a (hyps: %s)\n%!" Kat.print_expr x Kat.print_expr y hyps

(* printing the dfa of an expression in DOT format *)
let print_dfa x =
  let x = hexpr (Parse.expr x) in
  let trace ~exclude = SDFA.trace ~exclude in
  Trace.clear();
  (match !construction with
    | `Antimirov -> 
      let a,f = SNFA.reindex (Antimirov.nfa()) in 
      trace ~exclude:IntSet.empty (Determinisation.optimised a) (f (Antimirov.split x))
    | `Brzozowski -> 
      let a = Brzozowski.dfa() in
      trace ~exclude:Kat.zer a x 
    | `IlieYu -> 
      let a,x,_ = IlieYu.enfa x Kat.zer in 
      let a = Determinisation.optimised (Epsilon.remove a) in
      trace ~exclude:IntSet.empty a (IntSet.singleton x));
  print_endline (Trace.render !hk "\"")

(* entry point, for standalone program *)
let main() =
  let args = Arg.align [
    "-q", Arg.Set quiet, " quiet mode";
    "-s", Arg.Set stats, " print statistics";
    "-ssf", Arg.Set ssf, " put expressions in strict star form first";
    "-hyp", Arg.Set_string hypotheses, "l exploit hypotheses (i.e., specific KAT equations) in l";
    "-ant", Arg.Unit (set `Antimirov), " use Antimirov' partial derivatives [default]";
    "-brz", Arg.Unit (set `Brzozowski), " use Brzozowski's derivatives";
    "-iyu", Arg.Unit (set `IlieYu), " use Ilie and Yu's construction";
    "-hk", Arg.Set hk, " use the symbolic version of Hopcroft and Karp's algorithm";
    "-con", Arg.Set congruence, " additionally use up-to congruence";
    "-trc", Arg.Set trace, " trace the execution of the algorithm, and print it in DOT format";
    "-dfa", Arg.String print_dfa, "x print the symbolic DFA computed for x, in DOT format";
    "-cat", Arg.String (fun s -> Format.printf "%a\n%!" Kat.print_expr' (hexpr (Parse.expr s))),
    "x print the symbolic version of the expression x"
  ] in
  let doc = Format.sprintf "usage: %s [options] [x y]

This programs checks equivalence of KAT expressions [x] and [y].
See http://perso.ens-lyon.fr/damien.pous/symbolickat/ for a complete description.

The syntax for KAT expressions is the following:
 - atomic tests are characters from `a' to `j'
 - atomic Kleene elements are characters from `k' to `z'
 - multiplication or Boolean conjunction is implicit, by juxtaposition
 - addition or Boolean disjunction is `+'
 - Kleene star is postfix `*'
 - Boolean negation is prefix `!'
 - zero and one are `0' and `1'
Typically, `(ap+q)*!a' is a wellformed expression.

The [-hyps l] option allows one to exploit KAT hypotheses when comparing the 
expressions for equivalence. When used, [l] has to be a semi-colon (;) 
separated list of equations of shape expr<expr, expr=expr, or expr>expr.
Only some kinds of equations can be eliminated: 
- x=0, for an arbitrary expression x;
- ax=xb, ax<xb, or ax>xb, for arbitrary tests a and b and expression x;
- ap=a, or pa=a, for arbitrary test a, but atomic variable p;
- p=q, for atomic variables p and q;
An exception is raised if a given equation is not recognised to belong to 
one of these cases.

Options are the following ones; each command (e.g., -cat, -dfa, ...) is 
executed using the options that precede it, not the ones that follow it:" Sys.argv.(0) in
  let usage() = Arg.usage args doc in
  let anon =
    let one = ref None in
    let two = ref false in
    let answer fmt = 
      if !quiet then Format.ifprintf Format.std_formatter fmt
      else if !trace then Format.printf ("//"^^fmt)
      else Format.printf fmt
    in
    fun y -> 
      match !one with None -> one := Some y | Some x ->
	if !two then (usage(); exit 2);
	two := true;
	let e = match snd (compare ~hyps:(Parse.hyps !hypotheses) (Parse.expr x) (Parse.expr y)) with
	  | `E -> 
	    answer "%s = %s\n" x y; 0
	  | `L g -> 
	    answer "%s < %s\n" x y; 
	    answer "(strict because %a does not belong to the lhs)\n" print_gstring g; 1
	  | `G g -> 
	    answer "%s > %s\n" x y; 
	    answer "(strict because %a does not belong to the rhs)\n" print_gstring g; 2
	  | `D (g,g') -> 
	    answer "%s ≠ %s\n" x y;
	    answer "(%a does not belong to the rhs, and %a does not belong to the lhs)\n"
	      print_gstring g print_gstring g'; 3
	in
	if !stats then Stats.print Format.std_formatter;
	if !trace then Format.printf "%s@." (Trace.render !hk "\"");
	exit e
  in Arg.parse args anon doc

let _ = 
  if not !Sys.interactive then 
  try main () with Hypotheses.Unusable s -> failwith s

